from unittest.mock import patch

from django.core.exceptions import ValidationError
from edc_constants.constants import (
    COMPLETE,
    ESTIMATED,
    MEASURED,
    NO,
    NOT_APPLICABLE,
    YES,
)
from edc_form_validators.tests.mixins import FormValidatorTestMixin

from intecomm_form_validators.subject import VitalsFormValidator as Base

from ..mock_models import VitalsMockModel
from ..test_case_mixin import TestCaseMixin


class VitalsFormValidator(FormValidatorTestMixin, Base):
    pass


class VitalsFormValidationTests(TestCaseMixin):
    def setUp(self) -> None:
        super().setUp()
        is_baseline_patcher = patch(
            "intecomm_form_validators.subject.vitals_form_validator.is_baseline"
        )
        self.addCleanup(is_baseline_patcher.stop)
        self.mock_is_baseline = is_baseline_patcher.start()

        is_end_of_study_patcher = patch(
            "intecomm_form_validators.subject.vitals_form_validator.is_end_of_study"
        )
        self.addCleanup(is_end_of_study_patcher.stop)
        self.mock_is_end_of_study = is_end_of_study_patcher.start()

        raise_missing_clinical_review_patcher = patch(
            "edc_dx_review.utils.raise_if_clinical_review_does_not_exist"
        )
        self.addCleanup(raise_missing_clinical_review_patcher.stop)
        self.raise_missing_clinical_review = raise_missing_clinical_review_patcher.start()

    def set_baseline(self):
        self.mock_is_baseline.return_value = True
        self.mock_is_end_of_study.return_value = False

    def set_midpoint(self):
        self.mock_is_baseline.return_value = False
        self.mock_is_end_of_study.return_value = False

    def set_end_of_study(self):
        self.mock_is_baseline.return_value = False
        self.mock_is_end_of_study.return_value = True

    def get_cleaned_data(self) -> dict:
        cleaned_data = dict(
            weight_determination=MEASURED,
            weight=60.0,
            height=182.88,
            temperature=37.5,
            bp_one_taken=YES,
            bp_one_not_taken_reason="",
            sys_blood_pressure_one=120,
            dia_blood_pressure_one=80,
            bp_two_taken=YES,
            bp_two_not_taken_reason="",
            sys_blood_pressure_two=119,
            dia_blood_pressure_two=79,
            severe_htn=NO,
            comments="",
            crf_status=COMPLETE,
            crf_status_comments="",
        )
        return cleaned_data

    def test_cleaned_data_at_baseline_ok(self):
        self.set_baseline()
        cleaned_data = self.get_cleaned_data()
        form_validator = VitalsFormValidator(cleaned_data=cleaned_data, model=VitalsMockModel)
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_cleaned_data_after_baseline_ok(self):
        self.set_midpoint()
        cleaned_data = self.get_cleaned_data()
        form_validator = VitalsFormValidator(cleaned_data=cleaned_data, model=VitalsMockModel)
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_cleaned_data_at_end_of_study_ok(self):
        self.set_end_of_study()
        cleaned_data = self.get_cleaned_data()
        form_validator = VitalsFormValidator(cleaned_data=cleaned_data, model=VitalsMockModel)
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_weight_determination_not_measured_at_baseline_raises(self):
        self.set_baseline()
        for weight_determination in [ESTIMATED, NOT_APPLICABLE]:
            with self.subTest(weight_determination=weight_determination):
                cleaned_data = self.get_cleaned_data()
                cleaned_data.update(
                    {
                        "weight_determination": weight_determination,
                        "weight": 60 if weight_determination == ESTIMATED else None,
                    }
                )
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                with self.assertRaises(ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("weight_determination", cm.exception.error_dict)
                self.assertIn(
                    "Expected weight to be measured at this timepoint",
                    str(cm.exception.error_dict.get("weight_determination")),
                )

    def test_weight_determination_not_measured_at_end_of_study_raises(self):
        self.set_end_of_study()
        for weight_determination in [ESTIMATED, NOT_APPLICABLE]:
            with self.subTest(weight_determination=weight_determination):
                cleaned_data = self.get_cleaned_data()
                cleaned_data.update(
                    {
                        "weight_determination": weight_determination,
                        "weight": 60 if weight_determination == ESTIMATED else None,
                    }
                )
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                with self.assertRaises(ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("weight_determination", cm.exception.error_dict)
                self.assertIn(
                    "Expected weight to be measured at this timepoint",
                    str(cm.exception.error_dict.get("weight_determination")),
                )

    def test_weight_determination_not_measured_at_midpoint_ok(self):
        self.set_midpoint()
        for weight_determination in [ESTIMATED, NOT_APPLICABLE]:
            with self.subTest(weight_determination=weight_determination):
                cleaned_data = self.get_cleaned_data()
                cleaned_data.update(
                    {
                        "weight_determination": weight_determination,
                        "weight": 60 if weight_determination == ESTIMATED else None,
                    }
                )
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                try:
                    form_validator.validate()
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_missing_weight_at_baseline_raises(self):
        self.set_baseline()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "weight_determination": MEASURED,
                "weight": None,
            }
        )
        form_validator = VitalsFormValidator(cleaned_data=cleaned_data, model=VitalsMockModel)
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertIn("weight", cm.exception.error_dict)
        self.assertIn(
            "This field is required",
            str(cm.exception.error_dict.get("weight")),
        )

    def test_missing_weight_after_baseline_ok(self):
        self.set_midpoint()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "weight_determination": NOT_APPLICABLE,
                "weight": None,
            }
        )
        form_validator = VitalsFormValidator(cleaned_data=cleaned_data, model=VitalsMockModel)
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_weight_estimated_after_baseline_ok(self):
        self.set_midpoint()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "weight_determination": ESTIMATED,
                "weight": 55,
            }
        )
        form_validator = VitalsFormValidator(cleaned_data=cleaned_data, model=VitalsMockModel)
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_missing_weight_at_end_of_study_raises(self):
        self.set_end_of_study()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update(
            {
                "weight_determination": MEASURED,
                "weight": None,
            }
        )
        form_validator = VitalsFormValidator(cleaned_data=cleaned_data, model=VitalsMockModel)
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertIn("weight", cm.exception.error_dict)
        self.assertIn(
            "This field is required",
            str(cm.exception.error_dict.get("weight")),
        )

    def test_missing_height_at_baseline_raises(self):
        self.set_baseline()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update({"height": None})
        form_validator = VitalsFormValidator(cleaned_data=cleaned_data, model=VitalsMockModel)
        with self.assertRaises(ValidationError) as cm:
            form_validator.validate()
        self.assertIn("height", cm.exception.error_dict)
        self.assertIn(
            "This field is required",
            str(cm.exception.error_dict.get("height")),
        )

    def test_missing_height_after_baseline_ok(self):
        self.set_midpoint()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update({"height": None})
        form_validator = VitalsFormValidator(cleaned_data=cleaned_data, model=VitalsMockModel)
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_missing_height_at_end_of_study_ok(self):
        self.set_end_of_study()
        cleaned_data = self.get_cleaned_data()
        cleaned_data.update({"height": None})
        form_validator = VitalsFormValidator(cleaned_data=cleaned_data, model=VitalsMockModel)
        try:
            form_validator.validate()
        except ValidationError as e:
            self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_sys_lt_dia_blood_pressure_raises(self):
        self.set_baseline()
        for reading in ["one", "two"]:
            for sys_bp, dia_bp in [(80, 120), (90, 100), (99, 100)]:
                with self.subTest(reading=reading, sys_bp=sys_bp, dia_bp=dia_bp):
                    cleaned_data = self.get_cleaned_data()
                    cleaned_data.update(
                        {
                            f"sys_blood_pressure_{reading}": sys_bp,
                            f"dia_blood_pressure_{reading}": dia_bp,
                            "severe_htn": NO,
                        }
                    )
                    form_validator = VitalsFormValidator(
                        cleaned_data=cleaned_data, model=VitalsMockModel
                    )
                    with self.assertRaises(ValidationError) as cm:
                        form_validator.validate()
                    self.assertIn(f"dia_blood_pressure_{reading}", cm.exception.error_dict)
                    self.assertIn(
                        "Invalid. Diastolic must be less than systolic.",
                        str(cm.exception.error_dict.get(f"dia_blood_pressure_{reading}")),
                    )

    def test_sys_blood_pressure_indicates_severe_htn(self):
        self.set_baseline()
        for reading_one, reading_two in [
            (179, 181),
            (181, 179),
            (180, 180),
            (200, 160),
            (160, 200),
            (300, 120),
            (120, 300),
        ]:
            with self.subTest(reading_one=reading_one, reading_two=reading_two):
                cleaned_data = self.get_cleaned_data()
                cleaned_data.update(
                    {
                        "sys_blood_pressure_one": reading_one,
                        "sys_blood_pressure_two": reading_two,
                        "severe_htn": NO,
                    }
                )
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                with self.assertRaises(ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("severe_htn", cm.exception.error_dict)
                self.assertIn(
                    "Invalid. Patient has severe hypertension",
                    str(cm.exception.error_dict.get("severe_htn")),
                )

                cleaned_data.update({"severe_htn": YES})
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                try:
                    form_validator.validate()
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_dia_blood_pressure_indicates_severe_htn(self):
        self.set_baseline()
        for reading_one, reading_two in [
            (109, 111),
            (111, 109),
            (110, 110),
            (150, 81),
            (81, 150),
            (179, 81),
            (81, 179),
        ]:
            with self.subTest(reading_one=reading_one, reading_two=reading_two):
                cleaned_data = self.get_cleaned_data()
                cleaned_data.update(
                    {
                        "sys_blood_pressure_one": 180,
                        "dia_blood_pressure_one": reading_one,
                        "sys_blood_pressure_two": 180,
                        "dia_blood_pressure_two": reading_two,
                        "severe_htn": NO,
                    }
                )
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                with self.assertRaises(ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("severe_htn", cm.exception.error_dict)
                self.assertIn(
                    "Invalid. Patient has severe hypertension",
                    str(cm.exception.error_dict.get("severe_htn")),
                )

                cleaned_data.update({"severe_htn": YES})
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                try:
                    form_validator.validate()
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_only_one_bp_reading_can_indicate_severe_htn(self):
        self.set_baseline()
        for sys_bp, dia_bp in [
            (180, 80),
            (181, 80),
            (180, 110),
            (181, 111),
            (120, 110),
            (120, 111),
        ]:
            with self.subTest(sys_bp=sys_bp, dia_bp=dia_bp):
                cleaned_data = self.get_cleaned_data()
                cleaned_data.update(
                    {
                        "bp_one_taken": YES,
                        "bp_one_not_taken_reason": "",
                        "sys_blood_pressure_one": sys_bp,
                        "dia_blood_pressure_one": dia_bp,
                        "bp_two_taken": NO,
                        "bp_two_not_taken_reason": "Machine failed",
                        "sys_blood_pressure_two": None,
                        "dia_blood_pressure_two": None,
                        "severe_htn": NO,
                    }
                )
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                with self.assertRaises(ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("severe_htn", cm.exception.error_dict)
                self.assertIn(
                    "Invalid. Patient has severe hypertension",
                    str(cm.exception.error_dict.get("severe_htn")),
                )

                cleaned_data.update({"severe_htn": YES})
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                try:
                    form_validator.validate()
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_severe_htn_applicable_if_bp_taken(self):
        for severe_htn in [YES, NO]:
            with self.subTest(severe_htn=severe_htn):
                cleaned_data = self.get_cleaned_data()
                cleaned_data.update(
                    {
                        "bp_one_taken": YES,
                        "sys_blood_pressure_one": 120,
                        "dia_blood_pressure_one": 80,
                        "bp_two_taken": NO,
                        "bp_two_not_taken_reason": "Machine broke",
                        "sys_blood_pressure_two": None,
                        "dia_blood_pressure_two": None,
                        "severe_htn": NOT_APPLICABLE,
                    }
                )
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                with self.assertRaises(ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("severe_htn", cm.exception.error_dict)
                self.assertIn(
                    "This field is applicable",
                    str(cm.exception.error_dict.get("severe_htn")),
                )

                cleaned_data.update({"severe_htn": YES})
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                try:
                    form_validator.validate()
                except ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")

    def test_severe_htn_not_applicable_if_bp_not_taken(self):
        for severe_htn in [YES, NO]:
            with self.subTest(severe_htn=severe_htn):
                cleaned_data = self.get_cleaned_data()
                cleaned_data.update(
                    {
                        "bp_one_taken": NO,
                        "bp_one_not_taken_reason": "Machine broken",
                        "sys_blood_pressure_one": None,
                        "dia_blood_pressure_one": None,
                        "bp_two_taken": NOT_APPLICABLE,
                        "bp_two_not_taken_reason": "",
                        "sys_blood_pressure_two": None,
                        "dia_blood_pressure_two": None,
                        "severe_htn": severe_htn,
                    }
                )
                form_validator = VitalsFormValidator(
                    cleaned_data=cleaned_data, model=VitalsMockModel
                )
                with self.assertRaises(ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("severe_htn", cm.exception.error_dict)
                self.assertIn(
                    "This field is not applicable",
                    str(cm.exception.error_dict.get("severe_htn")),
                )



from edc_crf.crf_form_validator_mixins import CrfFormValidatorMixin
from edc_dx_review.utils import raise_if_clinical_review_does_not_exist
from edc_form_validators import FormValidator

from .mixins import BPFormValidatorMixin


class HtnReviewFormValidator(BPFormValidatorMixin, CrfFormValidatorMixin, FormValidator):
    def clean(self):
        raise_if_clinical_review_does_not_exist(self.cleaned_data.get("subject_visit"))
        self.validate_bp_reading(
            "sys_blood_pressure",
            "dia_blood_pressure",
        )



from __future__ import annotations

from django.test import TestCase
from django_mock_queries.query import MockSet
from edc_constants.constants import DM, HIV, HTN, NO, YES

from .mock_models import (
    AppointmentMockModel,
    ConditionsMockModel,
    PatientLogMockModel,
    SubjectVisitMockModel,
)


class TestCaseMixin(TestCase):
    def get_mock_patients(
        self,
        dm: int = None,
        htn: int = None,
        hiv: int = None,
        ncd: int = None,
        hiv_ncd: int = None,
        stable: bool | None = None,
        screen: bool | None = None,
        consent: bool | None = None,
        site: str | None = None,
    ) -> list:
        """Returns a list of mock patient logs"""
        patients = []
        default_ratio = (5, 5, 4, 0, 0)
        ratio = (dm or 0, htn or 0, hiv or 0, ncd or 0, hiv_ncd or 0) or default_ratio
        for i in range(0, ratio[0]):
            patients.append(
                self.get_mock_patient(
                    DM,
                    i=i + 100,
                    stable=stable,
                    screen=screen,
                    consent=consent,
                    site=site,
                )
            )
        for i in range(0, ratio[1]):
            patients.append(
                self.get_mock_patient(
                    HTN,
                    i=i + 200,
                    stable=stable,
                    screen=screen,
                    consent=consent,
                    site=site,
                )
            )
        for i in range(0, ratio[2]):
            patients.append(
                self.get_mock_patient(
                    HIV,
                    i=i + 300,
                    stable=stable,
                    screen=screen,
                    consent=consent,
                    site=site,
                )
            )
        for i in range(0, ratio[3]):
            patients.append(
                self.get_mock_patient(
                    DM,
                    HTN,
                    i=i + 300,
                    stable=stable,
                    screen=screen,
                    consent=consent,
                    site=site,
                )
            )
        for i in range(0, ratio[4]):
            patients.append(
                self.get_mock_patient(
                    HIV,
                    DM,
                    HTN,
                    i=i + 300,
                    stable=stable,
                    screen=screen,
                    consent=consent,
                    site=site,
                )
            )
        return patients

    @staticmethod
    def get_mock_patient(
        *conditions: str | list[str],
        i: int | None = None,
        stable: bool | None = None,
        screen: bool | None = None,
        consent: bool | None = None,
        site: str | None = None,
    ):
        """Returns a mock patient log"""
        # conditions = [condition] if isinstance(condition, (str,)) else condition
        stable = YES if stable else NO
        screening_identifier = f"XYZ{str(i)}" if screen else None
        subject_identifier = f"999-{str(i)}" if consent else None
        return PatientLogMockModel(
            name=f"NAME-{str(i)}",
            stable=stable,
            screening_identifier=screening_identifier,
            subject_identifier=subject_identifier,
            conditions=MockSet(
                *[ConditionsMockModel(name=x) for x in conditions],
            ),
            site=site,
        )

    @staticmethod
    def get_subject_visit(
        schedule_name: str = None,
        visit_code: str = None,
        visit_code_sequence: int = None,
        timepoint: int = None,
    ):
        appointment = AppointmentMockModel(
            schedule_name=schedule_name,
            visit_code=visit_code,
            visit_code_sequence=visit_code_sequence,
            timepoint=timepoint,
        )
        return SubjectVisitMockModel(appointment=appointment)



from edc_constants.constants import YES
from edc_crf.crf_form_validator import CrfFormValidator
from edc_visit_schedule.utils import raise_if_baseline


class ClinicalReviewFormValidator(CrfFormValidator):
    def clean(self):
        raise_if_baseline(self.cleaned_data.get("subject_visit"))
        self.required_if(
            YES,
            field="health_insurance",
            field_required="health_insurance_monthly_pay",
        )
        self.required_if(
            YES,
            field="patient_club",
            field_required="patient_club_monthly_pay",
        )



from unittest.mock import patch

from dateutil.relativedelta import relativedelta
from django import forms
from django_mock_queries.query import MockModel, MockSet
from edc_constants.constants import FEMALE, MALE
from edc_utils import get_utcnow

from intecomm_form_validators.screening import PatientLogFormValidator as Base

from ..mock_models import PatientGroupMockModel, PatientLogMockModel
from ..test_case_mixin import TestCaseMixin


class PatientLogTests(TestCaseMixin):
    @staticmethod
    def get_form_validator_cls(subject_screening=None):
        class PatientLogFormValidator(Base):
            @property
            def subject_screening(self):
                return subject_screening

        return PatientLogFormValidator

    def test_raises_if_last_appt_date_is_future(self):
        patient_group = PatientGroupMockModel(name="PARKSIDE", randomized=None)
        patient_log = PatientLogMockModel()
        cleaned_data = dict(
            name="ERIK",
            patient_group=patient_group,
            report_datetime=get_utcnow(),
            last_appt_date=(get_utcnow() + relativedelta(days=30)).date(),
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data, instance=patient_log, model=PatientLogMockModel
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("last_appt_date", cm.exception.error_dict)
        cleaned_data.update(last_appt_date=(get_utcnow() - relativedelta(days=30)).date())
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data, instance=patient_log, model=PatientLogMockModel
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_raises_if_next_appt_date_is_past(self):
        patient_group = PatientGroupMockModel(name="PARKSIDE", randomized=None)
        patient_log = PatientLogMockModel()
        cleaned_data = dict(
            name="ERIK",
            patient_group=patient_group,
            report_datetime=get_utcnow(),
            next_appt_date=(get_utcnow() - relativedelta(days=30)).date(),
        )
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data, instance=patient_log, model=PatientLogMockModel
        )
        with self.assertRaises(forms.ValidationError) as cm:
            form_validator.validate()
        self.assertIn("next_appt_date", cm.exception.error_dict)

        cleaned_data.update(next_appt_date=(get_utcnow() + relativedelta(days=30)).date())
        form_validator = self.get_form_validator_cls()(
            cleaned_data=cleaned_data, instance=patient_log, model=PatientLogMockModel
        )
        try:
            form_validator.validate()
        except forms.ValidationError:
            self.fail("ValidationError unexpectedly raised")

    def test_patient_log_matches_screening(self):
        patient_log = PatientLogMockModel(name="ERIK")
        patient_group = PatientGroupMockModel(
            name="PARKSIDE", patients=MockSet(patient_log), randomized=None
        )
        data = [
            ("gender", "gender", MALE, MALE, "Gender", False),
            ("gender", "gender", FEMALE, MALE, "Gender", True),
            ("initials", "initials", "XX", "XX", "Initials", False),
            ("initials", "initials", "XX", "YY", "Initials", True),
            (
                "hospital_identifier",
                "hospital_identifier",
                "12345",
                "12345",
                "Identifier",
                False,
            ),
            (
                "hospital_identifier",
                "hospital_identifier",
                "12345",
                "54321",
                "Identifier",
                True,
            ),
            (
                "site",
                "site",
                MockModel(mock_name="Site", id=110),
                MockModel(mock_name="Site", id=110),
                "Site",
                False,
            ),
            (
                "site",
                "site",
                MockModel(mock_name="Site", id=110),
                MockModel(mock_name="Site", id=120),
                "Site",
                True,
            ),
        ]
        for values in data:
            screening_fld, log_fld, screening_value, log_value, word, should_raise = values
            with self.subTest(
                screening_fld=screening_fld,
                log_fld=log_fld,
                screening_value=screening_value,
                log_value=log_value,
                word=word,
                should_raise=should_raise,
            ):
                subject_screening = MockModel(
                    mock_name="SubjectScreening", **{screening_fld: screening_value}
                )
                cleaned_data = dict(
                    name="ERIK",
                    report_datetime=get_utcnow(),
                    patient_group=patient_group,
                    **{log_fld: log_value},
                )
                form_validator = self.get_form_validator_cls(subject_screening)(
                    cleaned_data=cleaned_data, instance=patient_log, model=PatientLogMockModel
                )
                if should_raise:
                    with self.assertRaises(forms.ValidationError) as cm:
                        form_validator.validate()
                    self.assertIn(word, str(cm.exception.error_dict.get("__all__")))
                else:
                    form_validator.validate()

    @patch(
        "intecomm_form_validators.screening.patient_log_form_validator"
        ".get_subject_screening_model_cls"
    )
    def test_get_subject_screening(self, mock_subject_screening_model_cls):
        form_validator = Base(
            cleaned_data={},
            instance=MockModel(mock_name="PatientLog", name="BUBBA", screening_identifier="B"),
        )
        self.assertIsNotNone(form_validator.subject_screening)

    # def test_attempt_to_change_patient_in_randomized_group_raises(self):
    #     patients = self.get_mock_patients(ratio=[10, 0, 4])
    #     save_patients = []
    #     for patient in patients:
    #         patient.id = uuid4()
    #         save_patients.append(patient)
    #     patient_group = PatientGroupMockModel(
    #         randomized=True, patients=MockSet(*save_patients)
    #     )
    #     patient_group.save()
    #
    #     patient_log = save_patients[0]
    #     cleaned_data = dict(
    #         next_appt_date=(get_utcnow() + relativedelta(days=60)).date(),
    #     )
    #     form_validator = self.get_form_validator_cls()(
    #         cleaned_data=cleaned_data, instance=patient_log, model=PatientLogMockModel
    #     )
    #     with self.assertRaises(forms.ValidationError) as cm:
    #         form_validator.validate()
    #     self.assertIn("next_appt_date", cm.exception.error_dict)

    def test_age_in_years_lt_18_raises(self):
        for age in [17, 15, 1, 0]:
            with self.subTest(age=age):
                patient_log = PatientLogMockModel()
                form_validator = self.get_form_validator_cls()(
                    cleaned_data={"age_in_years": age},
                    instance=patient_log,
                    model=PatientLogMockModel,
                )

                with self.assertRaises(forms.ValidationError) as cm:
                    form_validator.validate()
                self.assertIn("age_in_years", cm.exception.error_dict)
                self.assertIn(
                    "Invalid. Patient must be 18 years or older",
                    str(cm.exception.error_dict.get("age_in_years")),
                )

    def test_age_in_years_gte_18_ok(self):
        for age in [18, 19, 29, 99]:
            with self.subTest(age=age):
                patient_log = PatientLogMockModel()
                form_validator = self.get_form_validator_cls()(
                    cleaned_data={"age_in_years": age},
                    instance=patient_log,
                    model=PatientLogMockModel,
                )
                try:
                    form_validator.validate()
                except forms.ValidationError as e:
                    self.fail(f"ValidationError unexpectedly raised. Got {e}")



from edc_constants.constants import NO, PATIENT, YES
from edc_form_validators import INVALID_ERROR, FormValidator


class PatientCallFormValidator(FormValidator):
    def clean(self):
        self.applicable_if(YES, field="answered", field_applicable="respondent")
        self.applicable_if(YES, field="answered", field_applicable="survival_status")
        if (
            self.cleaned_data.get("respondent") == PATIENT
            and self.cleaned_data.get("survival_status") != YES
        ):
            self.raise_validation_error(
                {"survival_status": "Invalid. Patient is the respondent"}, INVALID_ERROR
            )
        self.applicable_if(YES, field="answered", field_applicable="catchment_area")
        if self.cleaned_data.get("respondent") == PATIENT and self.cleaned_data.get(
            "catchment_area"
        ) not in [YES, NO]:
            self.raise_validation_error(
                {"catchment_area": "Invalid. Patient is the respondent"}, INVALID_ERROR
            )



from django.core.exceptions import ObjectDoesNotExist
from edc_constants.constants import YES
from edc_form_validators import FormValidator
from edc_form_validators.base_form_validator import INVALID_ERROR
from edc_screening.utils import get_subject_screening_model_cls

from .patient_group_form_validator import INVALID_RANDOMIZE

INVALID_APPOINTMENT_DATE = "INVALID_APPOINTMENT_DATE"
INVALID_GROUP = "INVALID_GROUP"
INVALID_CHANGE_ALREADY_SCREENED = "INVALID_CHANGE_ALREADY_SCREENED"


class PatientLogFormValidator(FormValidator):
    def __init__(self, subject_screening=None, **kwargs) -> None:
        self._subject_screening = subject_screening
        super().__init__(**kwargs)

    @property
    def age_in_years(self) -> int | None:
        return self.cleaned_data.get("age_in_years")

    def clean(self):
        if (
            self.instance.id
            and self.instance.patientgroup_set.filter(randomized=True).exists()
        ):
            self.raise_validation_error(
                "A patient in a randomized group may not be changed", INVALID_RANDOMIZE
            )

        if self.subject_screening and self.subject_screening.gender != self.cleaned_data.get(
            "gender"
        ):
            self.raise_validation_error(
                "Patient has already screened. Gender may not change",
                INVALID_CHANGE_ALREADY_SCREENED,
            )
        if (
            self.subject_screening
            and self.subject_screening.initials != self.cleaned_data.get("initials")
        ):
            self.raise_validation_error(
                "Patient has already screened. Initials may not change",
                INVALID_CHANGE_ALREADY_SCREENED,
            )
        if (
            self.subject_screening
            and self.subject_screening.hospital_identifier
            != self.cleaned_data.get("hospital_identifier")
        ):
            self.raise_validation_error(
                "Patient has already screened. Heath Facility Identifier may not change",
                INVALID_CHANGE_ALREADY_SCREENED,
            )

        self.validate_age()

        if (
            self.subject_screening
            and self.cleaned_data.get("site")
            and self.subject_screening.site.id != self.cleaned_data.get("site").id
        ):
            self.raise_validation_error(
                "Patient has already screened. Site / Health Facility may not change",
                INVALID_CHANGE_ALREADY_SCREENED,
            )

        if (
            self.cleaned_data.get("last_appt_date")
            and self.cleaned_data.get("report_datetime")
            and self.cleaned_data.get("last_appt_date")
            > self.cleaned_data.get("report_datetime").date()
        ):
            self.raise_validation_error(
                {"last_appt_date": "Cannot be a future date"}, INVALID_APPOINTMENT_DATE
            )

        if (
            self.cleaned_data.get("next_appt_date")
            and self.cleaned_data.get("report_datetime")
            and self.cleaned_data.get("next_appt_date")
            < self.cleaned_data.get("report_datetime").date()
        ):
            self.raise_validation_error(
                {"next_appt_date": "Must be a future date"}, INVALID_APPOINTMENT_DATE
            )
        self.required_if(
            YES, field="first_health_talk", field_required="first_health_talk_date"
        )
        self.required_if(
            YES, field="second_health_talk", field_required="second_health_talk_date"
        )
        # self.validate_group_changes()

    # def validate_group_changes(self):
    #     if from_group := self.instance.patient_group:
    #         to_group = self.cleaned_data.get("patient_group")
    #         if not to_group:
    #             if from_group.status == COMPLETE:
    #                 self.raise_validation_error(
    #                     "Cannot remove from current group. Group is complete.", INVALID_GROUP
    #                 )
    #         elif to_group:
    #             if from_group.name == to_group.name:
    #                 pass
    #             elif from_group.status == COMPLETE:
    #                 self.raise_validation_error(
    #                     "Cannot remove from current group. Group is complete.", INVALID_GROUP
    #                 )
    #             elif to_group.status == COMPLETE:
    #                 self.raise_validation_error(
    #                     "Cannot add to group. Group is complete.", INVALID_GROUP
    #                 )

    @property
    def subject_screening(self):
        if not self._subject_screening:
            try:
                self._subject_screening = get_subject_screening_model_cls().objects.get(
                    screening_identifier=self.instance.screening_identifier
                )
            except ObjectDoesNotExist:
                self._subject_screening = None
        return self._subject_screening

    def validate_age(self) -> None:
        if self.age_in_years is not None and not (18 <= self.age_in_years < 110):
            self.raise_validation_error(
                {"age_in_years": "Invalid. Patient must be 18 years or older"},
                INVALID_ERROR,
            )
